/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

// Module used by test_import_module.js

export const MODULE_IMPORTED = true;

export const MODULE_URI = import.meta.url;
